from geoalchemy2 import WKBElement
from typing import Optional, List
from vortex.Tuple import Tuple, addTupleType, TupleField

from peek_plugin_diagram._private.PluginNames import diagramTuplePrefix
from peek_plugin_diagram.tuples.model.ImportLiveDbDispLinkTuple import \
    ImportLiveDbDispLinkTuple
from peek_plugin_diagram.tuples.shapes.ImportDispPolylineTuple import \
    ImportDispPolylineTuple


@addTupleType
class ImportDispEdgeTemplateTuple(Tuple):
    """ Imported Shape Line Template

    This tuple is used by other plugins to load Line Template into the diagram.
    The line template is used by the user to create new lines.

    """
    __tupleType__ = diagramTuplePrefix + 'ImportDispEdgeTemplateTuple'

    # The size of the end types are relative to the line width
    END_TYPE_NONE = ImportDispPolylineTuple.END_TYPE_NONE
    END_TYPE_ARROW = ImportDispPolylineTuple.END_TYPE_ARROW
    END_TYPE_DOT = ImportDispPolylineTuple.END_TYPE_DOT

    ### BEGIN DISP COMMON FIELDS ###

    #: Key, This value is a unique ID of the object that this graphic represents
    # It's used to link this graphical object to objects in other plugins, like vertices
    # in the peek-plugin-graphdb plugin.
    # Length = 50
    key: str = TupleField()

    #: Selectable, Is is this item selectable?, the layer also needs selectable=true
    selectable: bool = TupleField()

    #: Data, Generic data, this is passed to the popup context in the UI.
    # peek_plugin_diagram doesn't care as long as it's json compatible or None
    # Json length Length = 400
    data: Optional[dict] = TupleField(None)

    #: The hash of the level to link to (Matches ImportDispLevel.importHash)
    levelHash: str = TupleField()

    #: The hash of the layer to link to (Matches ImportDispLayer.importHash)
    layerHash: str = TupleField()

    #: The Z Order of this display object when compared against other objects on
    # same layer and level.
    zOrder: int = TupleField()

    #: The unique hash of this display object
    importHash: str = TupleField()

    #: The unique hash for all the display items imported in a group with this one.
    #: for example, a page or tile reference.
    importGroupHash: str = TupleField()

    #: The key of the ModelSet to import into
    modelSetKey: str = TupleField()

    #: The Key of the Coordinate Set to import into
    coordSetKey: str = TupleField()

    #: Related links to LiveDB values for this display item
    liveDbDispLinks: List[ImportLiveDbDispLinkTuple] = TupleField()

    ### BEGIN FIELDS FOR THIS DISP ###

    #: A name for this dispGroup
    name: str = TupleField()

    lineWidth: int = TupleField()
    lineStyleHash: str = TupleField()
    lineColorHash: Optional[str] = TupleField()

    geom: List[float] = TupleField()

    #: Start end type, is this an arrow, etc?
    startEndType: Optional[int] = TupleField()

    #: End End Type, See Start end type
    endEndType: Optional[int] = TupleField()



import logging
from datetime import datetime
from typing import List

import pytz
from sqlalchemy.sql.expression import asc
from twisted.internet import task, reactor, defer
from twisted.internet.defer import inlineCallbacks
from vortex.DeferUtil import deferToThreadWrapWithLogger, vortexLogFailure

from peek_plugin_diagram._private.server.controller.StatusController import \
    StatusController
from peek_plugin_diagram._private.storage.GridKeyIndex import \
    DispIndexerQueue as DispIndexerQueueTable

logger = logging.getLogger(__name__)


class DispCompilerQueueController:
    """ Grid Compiler

    Compile the disp items into the grid data

    1) Query for queue
    2) Process queue
    3) Delete from queue
    """

    ITEMS_PER_TASK = 500
    PERIOD = 0.200

    QUEUE_MAX = 20
    QUEUE_MIN = 4

    TASK_TIMEOUT = 60.0

    def __init__(self, ormSessionCreator, statusController: StatusController):
        self._dbSessionCreator = ormSessionCreator
        self._statusController: StatusController = statusController

        self._pollLoopingCall = task.LoopingCall(self._poll)
        self._lastQueueId = 0
        self._queueCount = 0

        self._chunksInProgress = set()

    def isBusy(self) -> bool:
        return self._queueCount != 0

    def start(self):
        self._statusController.setDisplayCompilerStatus(True, self._queueCount)
        d = self._pollLoopingCall.start(self.PERIOD, now=False)
        d.addCallbacks(self._timerCallback, self._timerErrback)

    def _timerErrback(self, failure):
        vortexLogFailure(failure, logger)
        self._statusController.setDisplayCompilerStatus(False, self._queueCount)
        self._statusController.setDisplayCompilerError(str(failure.value))

    def _timerCallback(self, _):
        self._statusController.setDisplayCompilerStatus(False, self._queueCount)

    def stop(self):
        if self._pollLoopingCall.running:
            self._pollLoopingCall.stop()

    def shutdown(self):
        self.stop()

    @inlineCallbacks
    def _poll(self):
        # We queue the grids in bursts, reducing the work we have to do.
        if self._queueCount > self.QUEUE_MIN:
            return

        queueItems = yield self._grabQueueChunk()

        if not queueItems:
            return

        for start in range(0, len(queueItems), self.ITEMS_PER_TASK):

            items = queueItems[start: start + self.ITEMS_PER_TASK]

            # If we're already processing these chunks, then return and try later
            if self._chunksInProgress & set([o.dispId for o in items]):
                return

            # This should never fail
            d = self._sendToWorker(items)
            d.addErrback(vortexLogFailure, logger)

            # Set the watermark
            self._lastQueueId = items[-1].id

            self._queueCount += 1
            if self._queueCount >= self.QUEUE_MAX:
                break

        self._statusController.setDisplayCompilerStatus(True, self._queueCount)


    @inlineCallbacks
    def _sendToWorker(self, items: List[DispIndexerQueueTable]):
        from peek_plugin_diagram._private.worker.tasks.DispCompilerTask import \
            compileDisps

        queueIds = [o.id for o in items]
        dispIds = list(set([o.dispId for o in items]))
        startTime = datetime.now(pytz.utc)

        # Add the chunks we're processing to the set
        self._chunksInProgress |= set([o.dispId for o in items])

        try:
            d = compileDisps.delay(queueIds, dispIds)
            d.addTimeout(self.TASK_TIMEOUT, reactor)

            yield d
            logger.debug("%s Disps, Time Taken = %s",
                         len(items), datetime.now(pytz.utc) - startTime)

            self._queueCount -= 1

            self._statusController.setDisplayCompilerStatus(True, self._queueCount)
            self._statusController.addToDisplayCompilerTotal(self.ITEMS_PER_TASK)

            # Success, Remove the chunks from the in-progress queue
            self._chunksInProgress -= set([o.dispId for o in items])

        except Exception as e:
            if isinstance(e, defer.TimeoutError):
                logger.info("Retrying compile, Task has timed out.")
            else:
                logger.debug("Retrying compile : %s", str(e))

            reactor.callLater(2.0, self._sendToWorker, items)
            return

    @deferToThreadWrapWithLogger(logger)
    def _grabQueueChunk(self):
        toGrab = (self.QUEUE_MAX - self._queueCount) * self.ITEMS_PER_TASK
        session = self._dbSessionCreator()
        try:
            queueItems = (session.query(DispIndexerQueueTable)
                          .order_by(asc(DispIndexerQueueTable.id))
                          .filter(DispIndexerQueueTable.id > self._lastQueueId)
                          .yield_per(toGrab)
                          .limit(toGrab)
                          .all())

            session.expunge_all()
            return queueItems
        finally:
            session.close()

    @deferToThreadWrapWithLogger(logger)
    def queueDisps(self, dispIds):
        return self.queueDispIdsToCompile(dispIds, self._dbSessionCreator)

    @classmethod
    def queueDispIdsToCompile(cls, dispIdsToCompile: List[int], ormSessionCreator):
        if not dispIdsToCompile:
            return

        ormSession = ormSessionCreator()
        try:
            cls.queueDispIdsToCompileWithSession(dispIdsToCompile, ormSession)
            ormSession.commit()

        finally:
            ormSession.close()

    @staticmethod
    def queueDispIdsToCompileWithSession(dispIdsToCompile: List[int], ormSessionOrConn):
        if not dispIdsToCompile:
            return

        logger.debug("Queueing %s disps for compile", len(dispIdsToCompile))

        inserts = []
        for dispId in dispIdsToCompile:
            inserts.append(dict(dispId=dispId))

        ormSessionOrConn.execute(DispIndexerQueueTable.__table__.insert(), inserts)

    @deferToThreadWrapWithLogger(logger)
    def _dedupeQueue(self):
        session = self._dbSessionCreator()
        try:
            session.execute("""
                 with sq_raw as (
                    SELECT "id", "dispId"
                    FROM pl_diagram."DispCompilerQueue"
                    WHERE id > %s
                    LIMIT %s
                ), sq as (
                    SELECT min(id) as "minId", "dispId"
                    FROM sq_raw
                    GROUP BY  "dispId"
                    HAVING count("dispId") > 1
                )
                DELETE
                FROM pl_diagram."DispCompilerQueue"
                     USING sq sq1
                WHERE pl_diagram."DispCompilerQueue"."id" != sq1."minId"
                    AND pl_diagram."DispCompilerQueue"."dispId" = sq1."dispId"
                    
            """ % (self._lastQueueId, self.QUEUE_MAX * self.ITEMS_PER_TASK))
            session.commit()
        finally:
            session.close()




import logging
from typing import List

from peek_plugin_base.PeekVortexUtil import peekServerName, peekClientName
from peek_plugin_diagram._private.PluginNames import diagramFilt
from peek_plugin_diagram._private.storage.GridKeyIndex import GridKeyIndexCompiled
from peek_plugin_diagram._private.tuples.grid.EncodedGridTuple import EncodedGridTuple
from peek_plugin_diagram._private.tuples.grid.GridTuple import GridTuple
from vortex.DeferUtil import vortexLogFailure
from vortex.rpc.RPC import vortexRPC

logger = logging.getLogger(__name__)


class ClientGridLoaderRpc:
    def __init__(self, liveDbWatchController, dbSessionCreator):
        self._liveDbWatchController = liveDbWatchController
        self._dbSessionCreator = dbSessionCreator

    def makeHandlers(self):
        """ Make Handlers

        In this method we start all the RPC handlers
        start() returns an instance of it's self so we can simply yield the result
        of the start method.

        """

        yield self.loadGrids.start(funcSelf=self)
        yield self.updateClientWatchedGrids.start(funcSelf=self)
        logger.debug("RPCs started")

    # -------------
    @vortexRPC(peekServerName, acceptOnlyFromVortex=peekClientName,timeoutSeconds=120,
               additionalFilt=diagramFilt, deferToThread=True)
    def loadGrids(self, offset: int, count: int) -> List[GridTuple]:
        """ Update Page Loader Status

        Tell the server of the latest status of the loader

        """
        session = self._dbSessionCreator()
        try:
            ormGrids = (session
                        .query(GridKeyIndexCompiled)
                        .order_by(GridKeyIndexCompiled.id)
                        .offset(offset)
                        .limit(count)
                        .yield_per(200))

            gridTuples: List[GridTuple] = []
            for ormGrid in ormGrids:
                gridTuples.append(
                    EncodedGridTuple(gridKey=ormGrid.gridKey,
                              encodedGridTuple=ormGrid.encodedGridTuple,
                              lastUpdate=ormGrid.lastUpdate)
                )

            return gridTuples

        finally:
            session.close()



    # -------------
    @vortexRPC(peekServerName, acceptOnlyFromVortex=peekClientName,
               additionalFilt=diagramFilt)
    def updateClientWatchedGrids(self, clientId: str, gridKeys: List[str]) -> None:
        """ Update Client Watched Grids

        Tell the server that these grids are currently being watched by users.

        :param clientId: A unique identifier of the client (Maybe it's vortex uuid)
        :param gridKeys: A list of grid keys that this client is observing.
        :returns: Nothing
        """

        d = self._liveDbWatchController.updateClientWatchedGrids(clientId, gridKeys)
        d.addErrback(vortexLogFailure, logger, consumeError=True)



"""removed more string sizes

Peek Plugin Database Migration Script

Revision ID: 473d5f6a97e8
Revises: 6f545d4166dc
Create Date: 2019-06-10 22:21:54.762566

"""

# revision identifiers, used by Alembic.
revision = '473d5f6a97e8'
down_revision = '6f545d4166dc'
branch_labels = None
depends_on = None

import sqlalchemy as sa
from alembic import op


def upgrade():
    op.alter_column('DispBase', 'importHash', type_=sa.String(), nullable=True,
                    schema='pl_diagram')
    op.alter_column('DispBase', 'importGroupHash', type_=sa.String(), nullable=True,
                    schema='pl_diagram')

    op.alter_column('LiveDbDispLink', 'dispAttrName', type_=sa.String(), nullable=False,
                    schema='pl_diagram')
    op.alter_column('LiveDbDispLink', 'liveDbKey', type_=sa.String(), nullable=False,
                    schema='pl_diagram')

    op.alter_column('LiveDbDispLink', 'importKeyHash', type_=sa.String(), nullable=True,
                    schema='pl_diagram')
    op.alter_column('LiveDbDispLink', 'importGroupHash', type_=sa.String(), nullable=True,
                    schema='pl_diagram')
    op.alter_column('LiveDbDispLink', 'importDispHash', type_=sa.String(), nullable=True,
                    schema='pl_diagram')
    op.alter_column('LiveDbDispLink', 'propsJson', type_=sa.String(), nullable=True,
                    schema='pl_diagram')

def downgrade():
    pass



from typing import List

from vortex.Tuple import Tuple, addTupleType, TupleField

from peek_plugin_diagram._private.PluginNames import diagramTuplePrefix


@addTupleType
class ImportBranchTuple(Tuple):
    """ Imported Branch

    This tuple is used by other plugins to load branches into the diagram.

    """
    __tupleType__ = diagramTuplePrefix + 'ImportBranchTuple'

    #:  The name of the model set for this branch
    modelSetKey: str = TupleField()

    #:  The Coordinate Set key that this branch applies to
    coordSetKey: str = TupleField()

    #:  The unique key for this branch
    key: str = TupleField()

    #:  The import hash for this branch
    importHash: str = TupleField()

    #:  The import hash for this branch
    importGroupHash: str = TupleField()

    #:  The alt color
    deltas: List = TupleField([])

    #:  The alt color
    addedDisps: List = TupleField([])

    #:  The alt color
    updatedDisps: List = TupleField([])

    #:  The alt color
    deletedDispKeys: List = TupleField([])

    #:  Is this branch Visible by default
    visible: bool = TupleField()



"""Increased size of disp.json

Peek Plugin Database Migration Script

Revision ID: cbd5e370686f
Revises: 9e7ca1acd6be
Create Date: 2017-08-12 17:34:00.532517

"""

# revision identifiers, used by Alembic.
revision = 'cbd5e370686f'
down_revision = '9e7ca1acd6be'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():

    op.alter_column('DispBase', 'dispJson',
               existing_type=sa.VARCHAR(length=200000),
               nullable=True,
               schema='pl_diagram')

    op.alter_column('DispPolygon', 'geomJson',
               existing_type=sa.VARCHAR(length=200000),
               nullable=False,
               schema='pl_diagram')

    op.alter_column('DispPolyline', 'geomJson',
               existing_type=sa.VARCHAR(length=200000),
               nullable=False,
               schema='pl_diagram')


def downgrade():
    raise Exception("Downgrade is not implemented")


import logging
from datetime import datetime
from typing import List

import pytz
from sqlalchemy import asc
from twisted.internet import task, reactor, defer
from twisted.internet.defer import inlineCallbacks
from vortex.DeferUtil import deferToThreadWrapWithLogger, vortexLogFailure

from peek_plugin_diagram._private.server.client_handlers.BranchIndexChunkUpdateHandler import \
    BranchIndexChunkUpdateHandler
from peek_plugin_diagram._private.server.controller.StatusController import \
    StatusController
from peek_plugin_diagram._private.storage.branch.BranchIndexCompilerQueue import \
    BranchIndexCompilerQueue

logger = logging.getLogger(__name__)


class BranchIndexCompilerQueueController:
    """ BranchIndexChunkCompilerQueueController

    Compile the disp items into the grid data

    1) Query for queue
    2) Process queue
    3) Delete from queue

    """

    ITEMS_PER_TASK = 10
    PERIOD = 1.000

    QUEUE_MAX = 20
    QUEUE_MIN = 0

    TASK_TIMEOUT = 60.0

    def __init__(self, dbSessionCreator,
                 statusController: StatusController,
                 clientUpdateHandler: BranchIndexChunkUpdateHandler):
        self._dbSessionCreator = dbSessionCreator
        self._statusController: StatusController = statusController
        self._clientUpdateHandler: BranchIndexChunkUpdateHandler = clientUpdateHandler

        self._pollLoopingCall = task.LoopingCall(self._poll)
        self._lastQueueId = -1
        self._queueCount = 0

        self._chunksInProgress = set()

    def start(self):
        self._statusController.setBranchIndexCompilerStatus(True, self._queueCount)
        d = self._pollLoopingCall.start(self.PERIOD, now=False)
        d.addCallbacks(self._timerCallback, self._timerErrback)

    def _timerErrback(self, failure):
        vortexLogFailure(failure, logger)
        self._statusController.setBranchIndexCompilerStatus(False, self._queueCount)
        self._statusController.setBranchIndexCompilerError(str(failure.value))

    def _timerCallback(self, _):
        self._statusController.setBranchIndexCompilerStatus(False, self._queueCount)

    def stop(self):
        if self._pollLoopingCall.running:
            self._pollLoopingCall.stop()

    def shutdown(self):
        self.stop()

    @inlineCallbacks
    def _poll(self):

        # We queue the grids in bursts, reducing the work we have to do.
        if self._queueCount > self.QUEUE_MIN:
            return

        # Check for queued items
        queueItems = yield self._grabQueueChunk()
        if not queueItems:
            return

        # Send the tasks to the peek worker
        for start in range(0, len(queueItems), self.ITEMS_PER_TASK):

            items = queueItems[start: start + self.ITEMS_PER_TASK]

            # If we're already processing these chunks, then return and try later
            if self._chunksInProgress & set([o.chunkKey for o in items]):
                return

            # Set the watermark
            self._lastQueueId = items[-1].id

            # This should never fail
            d = self._sendToWorker(items)
            d.addErrback(vortexLogFailure, logger)

            self._queueCount += 1
            if self._queueCount >= self.QUEUE_MAX:
                break

        yield self._dedupeQueue()

    @inlineCallbacks
    def _sendToWorker(self, items: List[BranchIndexCompilerQueue]):
        from peek_plugin_diagram._private.worker.tasks.branch.BranchIndexCompiler import \
            compileBranchIndexChunk

        startTime = datetime.now(pytz.utc)

        # Add the chunks we're processing to the set
        self._chunksInProgress |= set([o.chunkKey for o in items])

        try:
            d = compileBranchIndexChunk.delay(items)
            d.addTimeout(self.TASK_TIMEOUT, reactor)

            chunkKeys = yield d
            logger.debug("Time Taken = %s" % (datetime.now(pytz.utc) - startTime))
            self._queueCount -= 1
            self._clientUpdateHandler.sendChunks(chunkKeys)
            self._statusController.addToBranchIndexCompilerTotal(len(items))
            self._statusController.setBranchIndexCompilerStatus(True, self._queueCount)

            # Success, Remove the chunks from the in-progress queue
            self._chunksInProgress -= set([o.chunkKey for o in items])

        except Exception as e:
            if isinstance(e, defer.TimeoutError):
                logger.warning("Retrying compile, Task has timed out.")
            else:
                logger.debug("Retrying compile : %s", str(e))

            reactor.callLater(2.0, self._sendToWorker, items)
            return

    @deferToThreadWrapWithLogger(logger)
    def _grabQueueChunk(self):
        session = self._dbSessionCreator()
        try:
            qry = (session.query(BranchIndexCompilerQueue)
                   .order_by(asc(BranchIndexCompilerQueue.id))
                   .filter(BranchIndexCompilerQueue.id > self._lastQueueId)
                   .yield_per(self.QUEUE_MAX)
                   .limit(self.QUEUE_MAX)
                   )

            queueItems = qry.all()
            session.expunge_all()

            # Deduplicate the values and return the ones with the lowest ID
            return list({o.chunkKey: o for o in reversed(queueItems)}.values())

        finally:
            session.close()

    @deferToThreadWrapWithLogger(logger)
    def _dedupeQueue(self):
        session = self._dbSessionCreator()
        try:
            session.execute("""
                with sq as (
                    SELECT min(id) as "minId"
                    FROM pl_diagram."BranchIndexCompilerQueue"
                    WHERE id > %s
                    GROUP BY "modelSetId", "chunkKey"
                )
                DELETE
                FROM pl_diagram."BranchIndexCompilerQueue"
                WHERE "id" not in (SELECT "minId" FROM sq)
            """ % self._lastQueueId)
            session.commit()
        finally:
            session.close()



"""added compiler q indexes

Peek Plugin Database Migration Script

Revision ID: 2fc6d3246717
Revises: 1856b7fb8803
Create Date: 2019-10-11 15:19:48.918248

"""

# revision identifiers, used by Alembic.
revision = '2fc6d3246717'
down_revision = '1856b7fb8803'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_index('idx_BICompQueue_modelSetId_chunkKey', 'BranchIndexCompilerQueue', ['modelSetId', 'chunkKey'], unique=False, schema='pl_diagram')
    op.create_index('idx_DispCompQueue_dispId', 'DispCompilerQueue', ['dispId'], unique=False, schema='pl_diagram')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index('idx_DispCompQueue_dispId', table_name='DispCompilerQueue', schema='pl_diagram')
    op.drop_index('idx_BICompQueue_modelSetId_chunkKey', table_name='BranchIndexCompilerQueue', schema='pl_diagram')
    # ### end Alembic commands ###


from vortex.Tuple import addTupleType, TupleField
from vortex.TupleAction import TupleActionABC

from peek_plugin_diagram._private.PluginNames import diagramTuplePrefix
from peek_plugin_diagram._private.tuples.branch.BranchTuple import BranchTuple


@addTupleType
class BranchLiveEditTupleAction(TupleActionABC):
    """ Branch Live Edit Tuple Action

    This is the Branch Live Edit tuple Action

    """
    __tupleType__ = diagramTuplePrefix + 'BranchLiveEditTupleAction'

    EDITING_STARTED = 1
    EDITING_UPDATED = 2
    EDITING_FINISHED = 3
    EDITING_SAVED = 3

    updatedByUser: str = TupleField()
    actionType: int = TupleField()
    branchTuple: BranchTuple = TupleField()



"""Added disp data

Peek Plugin Database Migration Script

Revision ID: d103b9c1eba9
Revises: 558bc7fc4d36
Create Date: 2017-09-23 14:58:37.938811

"""

# revision identifiers, used by Alembic.
revision = 'd103b9c1eba9'
down_revision = '558bc7fc4d36'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    op.add_column('DispBase', sa.Column('dataJson', sa.String(length=500), nullable=True), schema='pl_diagram')


def downgrade():
    op.drop_column('DispBase', 'dataJson', schema='pl_diagram')


"""removed string field sizes

Peek Plugin Database Migration Script

Revision ID: 1b17ef0cca17
Revises: f76791d27f6d
Create Date: 2019-02-15 22:54:38.653794

"""

# revision identifiers, used by Alembic.
revision = '1b17ef0cca17'
down_revision = 'f76791d27f6d'
branch_labels = None
depends_on = None

import sqlalchemy as sa
from alembic import op


def upgrade():
    op.alter_column('DispBase', 'dispJson', type_=sa.String(), nullable=True,
                    schema='pl_diagram')
    op.alter_column('DispBase', 'locationJson', type_=sa.String(), nullable=True,
                    schema='pl_diagram')
    op.alter_column('DispBase', 'key', type_=sa.String(), nullable=True,
                    schema='pl_diagram')
    op.alter_column('DispBase', 'dataJson', type_=sa.String(), nullable=True,
                    schema='pl_diagram')

    op.alter_column('DispGroup', 'name', type_=sa.String(), nullable=False,
                    schema='pl_diagram')
    op.alter_column('DispText', 'geomJson', type_=sa.String(), nullable=False,
                    schema='pl_diagram')
    op.alter_column('DispPolygon', 'geomJson', type_=sa.String(), nullable=False,
                    schema='pl_diagram')
    op.alter_column('DispPolyline', 'geomJson', type_=sa.String(), nullable=False,
                    schema='pl_diagram')
    op.alter_column('DispEllipse', 'geomJson', type_=sa.String(), nullable=False,
                    schema='pl_diagram')
    op.alter_column('DispGroupPointer', 'geomJson', type_=sa.String(), nullable=False,
                    schema='pl_diagram')


def downgrade():
    op.alter_column('DispBase', 'dispJson', type_=sa.String(length=200000), nullable=True,
                    schema='pl_diagram')
    op.alter_column('DispBase', 'locationJson', type_=sa.String(length=120), nullable=True,
                    schema='pl_diagram')
    op.alter_column('DispBase', 'key', type_=sa.String(length=50), nullable=True,
                    schema='pl_diagram')
    op.alter_column('DispBase', 'dataJson', type_=sa.String(length=500), nullable=True,
                    schema='pl_diagram')

    op.alter_column('DispGroup', 'name', type_=sa.String(length=50), nullable=False,
                    schema='pl_diagram')
    op.alter_column('DispText', 'geomJson', type_=sa.String(length=2000), nullable=False,
                    schema='pl_diagram')
    op.alter_column('DispPolygon', 'geomJson', type_=sa.String(length=200000),
                    nullable=False, schema='pl_diagram')
    op.alter_column('DispPolyline', 'geomJson', type_=sa.String(length=200000),
                    nullable=False, schema='pl_diagram')
    op.alter_column('DispEllipse', 'geomJson', type_=sa.String(length=2000),
                    nullable=False, schema='pl_diagram')
    op.alter_column('DispGroupPointer', 'geomJson', type_=sa.String(length=2000),
                    nullable=False, schema='pl_diagram')



"""added disp zOrder

Peek Plugin Database Migration Script

Revision ID: c88ba8f6761f
Revises: 4925c894757b
Create Date: 2017-11-26 12:52:22.016867

"""

# revision identifiers, used by Alembic.
revision = 'c88ba8f6761f'
down_revision = '4925c894757b'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('DispBase', sa.Column('zOrder', sa.Integer(), server_default='0', nullable=True), schema='pl_diagram')
    op.execute(''' UPDATE "pl_diagram"."DispBase" SET "zOrder" = 0 ''')
    op.alter_column('DispBase', 'zOrder', type_=sa.Integer, nullable=False, schema='pl_diagram')


def downgrade():
    op.drop_column('DispBase', 'zOrder', schema='pl_diagram')



import logging

from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Integer, String
from sqlalchemy.orm import relationship
from sqlalchemy.sql.schema import Index, Sequence
from sqlalchemy.sql.sqltypes import DateTime

from peek_plugin_base.storage.TypeDecorators import PeekLargeBinary
from peek_plugin_diagram._private.PluginNames import diagramTuplePrefix
from vortex.Tuple import Tuple, addTupleType
from .DeclarativeBase import DeclarativeBase
from .Display import DispBase
from .ModelSet import ModelCoordSet

logger = logging.getLogger(__name__)


class DispIndexerQueue(DeclarativeBase):
    __tablename__ = 'DispCompilerQueue'

    id = Column(Integer, primary_key=True, autoincrement=True)
    dispId = Column(Integer, primary_key=True)

    __table_args__ = (
        Index("idx_DispCompQueue_dispId", dispId, unique=False),
    )


@addTupleType
class GridKeyCompilerQueue(Tuple, DeclarativeBase):
    __tablename__ = 'GridKeyCompilerQueue'
    __tupleType__ = diagramTuplePrefix + __tablename__

    id = Column(Integer, primary_key=True, autoincrement=True)

    gridKey = Column(String(30), primary_key=True)
    coordSetId = Column(Integer,
                        ForeignKey('ModelCoordSet.id', ondelete='CASCADE'),
                        primary_key=True)

    __table_args__ = (
        Index("idx_GKCompQueue_coordSetId_gridKey", coordSetId, gridKey, unique=False),
    )


@addTupleType
class GridKeyIndex(Tuple, DeclarativeBase):
    __tablename__ = 'GridKeyIndex'
    __tupleType__ = diagramTuplePrefix + __tablename__

    gridKey = Column(String(30), primary_key=True)
    dispId = Column(Integer,
                    ForeignKey('DispBase.id', ondelete='CASCADE'),
                    primary_key=True)

    disp = relationship(DispBase)

    coordSetId = Column(Integer, ForeignKey('ModelCoordSet.id', ondelete="CASCADE"), nullable=False)
    coordSet = relationship(ModelCoordSet)

    __table_args__ = (
        Index("idx_GridKeyIndex_gridKey", gridKey, unique=False),
        Index("idx_GridKeyIndex_dispId", dispId, unique=False),
        Index("idx_GridKeyIndex_coordSetId", coordSetId, unique=False),
    )


@addTupleType
class GridKeyIndexCompiled(Tuple, DeclarativeBase):
    __tablename__ = 'GridKeyIndexCompiled'
    __tupleType__ = diagramTuplePrefix + __tablename__

    id = Column(Integer, primary_key=True, autoincrement=True)

    gridKey = Column(String(30), nullable=False)
    encodedGridTuple = Column(PeekLargeBinary, nullable=False)
    lastUpdate = Column(String(50), nullable=False)

    coordSetId = Column(Integer,
                        ForeignKey('ModelCoordSet.id', ondelete='CASCADE'),
                        nullable=False)
    coordSet = relationship(ModelCoordSet)

    __table_args__ = (
        Index("idx_GKIndexUpdate_coordSetId", coordSetId, unique=False),
        Index("idx_GKIndexUpdate_gridKey", gridKey, unique=True),
    )



import hashlib
import logging
from _collections import defaultdict
from base64 import b64encode
from collections import namedtuple
from datetime import datetime
from functools import cmp_to_key
from typing import List

import pytz
from peek_plugin_base.worker import CeleryDbConn
from peek_plugin_diagram._private.storage.Display import DispLevel, DispBase, DispLayer
from peek_plugin_diagram._private.storage.GridKeyIndex import GridKeyIndexCompiled, \
    GridKeyCompilerQueue, \
    GridKeyIndex
from peek_plugin_diagram._private.tuples.grid.GridTuple import GridTuple
from peek_plugin_base.worker.CeleryApp import celeryApp
from txcelery.defer import DeferrableTask
from vortex.Payload import Payload

logger = logging.getLogger(__name__)

DispData = namedtuple('DispData', ['json', 'id', 'zOrder', 'levelOrder', 'layerOrder'])

""" Grid Compiler

Compile the disp items into the grid data

1) Query for queue
2) Process queue
3) Delete from queue
"""


@DeferrableTask
@celeryApp.task(bind=True)
def compileGrids(self, queueItems) -> List[str]:
    """ Compile Grids Task

    :param self: A celery reference to this task
    :param queueItems: An encoded payload containing the queue tuples.
    :returns: A list of grid keys that have been updated.
    """
    gridKeys = list(set([i.gridKey for i in queueItems]))
    coordSetIdByGridKey = {i.gridKey: i.coordSetId for i in queueItems}

    queueTable = GridKeyCompilerQueue.__table__
    gridTable = GridKeyIndexCompiled.__table__

    startTime = datetime.now(pytz.utc)

    session = CeleryDbConn.getDbSession()
    engine = CeleryDbConn.getDbEngine()
    conn = engine.connect()
    transaction = conn.begin()
    try:

        logger.debug("Staring compile of %s queueItems in %s",
                    len(queueItems), (datetime.now(pytz.utc) - startTime))

        total = 0
        dispData = _qryDispData(session, gridKeys)

        conn.execute(gridTable.delete(gridTable.c.gridKey.in_(gridKeys)))

        transaction.commit()
        transaction = conn.begin()

        inserts = []
        for gridKey, dispJsonStr in dispData.items():
            m = hashlib.sha256()
            m.update(gridKey.encode())
            m.update(dispJsonStr.encode())
            gridTupleHash = b64encode(m.digest()).decode()

            gridTuple = GridTuple(
                gridKey=gridKey,
                dispJsonStr=dispJsonStr,
                lastUpdate=gridTupleHash
            )

            encodedGridTuple = Payload(tuples=[gridTuple]).toEncodedPayload()

            inserts.append(dict(coordSetId=coordSetIdByGridKey[gridKey],
                                gridKey=gridKey,
                                lastUpdate=gridTupleHash,
                                encodedGridTuple=encodedGridTuple))

        if inserts:
            conn.execute(gridTable.insert(), inserts)

        logger.debug("Compiled %s gridKeys, %s missing, in %s",
                    len(inserts),
                    len(gridKeys) - len(inserts), (datetime.now(pytz.utc) - startTime))

        total += len(inserts)

        queueItemIds = [o.id for o in queueItems]
        conn.execute(queueTable.delete(queueTable.c.id.in_(queueItemIds)))

        transaction.commit()
        logger.info("Compiled and Committed %s GridKeyIndexCompileds in %s",
                    total, (datetime.now(pytz.utc) - startTime))

        return gridKeys

    except Exception as e:
        transaction.rollback()
        logger.debug("Compile of grids failed, retrying : %s", gridKeys)
        raise self.retry(exc=e, countdown=2)

    finally:
        conn.close()
        session.close()


def _dispBaseSortCmp(dispData1, dispData2):
    isData1None = dispData1.levelOrder is None or dispData1.levelOrder is None
    isData2None = dispData2.levelOrder is None or dispData2.levelOrder is None

    if isData1None and isData2None:
        return 0

    elif isData1None:
        return -1

    elif isData2None:
        return 1

    levelDiff = dispData1.levelOrder - dispData2.levelOrder
    if levelDiff != 0:
        return levelDiff

    layerDiff = dispData1.layerOrder - dispData2.layerOrder
    if layerDiff != 0:
        return layerDiff

    return dispData1.zOrder - dispData2.zOrder


def _qryDispData(session, gridKeys):
    indexQry = (
        session.query(GridKeyIndex.gridKey, DispBase.dispJson,
                      DispBase.id, DispBase.zOrder,
                      DispLevel.order, DispLayer.order)
            .join(DispBase, DispBase.id == GridKeyIndex.dispId)
            .outerjoin(DispLevel)
            .outerjoin(DispLayer)
            .filter(GridKeyIndex.gridKey.in_(gridKeys))
    )

    dispsByGridKeys = defaultdict(list)
    dispJsonByGridKey = {}

    for item in indexQry:
        dispsByGridKeys[item[0]].append(DispData(*item[1:]))

    for gridKey, dispDatas in list(dispsByGridKeys.items()):
        dispsDumpedJson = [
            d.json
            for d in sorted(dispDatas, key=cmp_to_key(_dispBaseSortCmp))
            if d.json
        ]

        dispJsonByGridKey[gridKey] = '[' + ','.join(dispsDumpedJson) + ']'

    return dispJsonByGridKey



import logging
from typing import Dict, List

from twisted.internet.defer import inlineCallbacks

from peek_plugin_diagram._private.PluginNames import diagramFilt
from peek_plugin_diagram._private.server.client_handlers.BranchIndexChunkLoadRpc import \
    BranchIndexChunkLoadRpc
from peek_plugin_diagram._private.storage.branch.BranchIndexEncodedChunk import \
    BranchIndexEncodedChunk
from vortex.PayloadEndpoint import PayloadEndpoint
from vortex.PayloadEnvelope import PayloadEnvelope

logger = logging.getLogger(__name__)

clientBranchIndexUpdateFromServerFilt = dict(key="clientBranchIndexUpdateFromServer")
clientBranchIndexUpdateFromServerFilt.update(diagramFilt)


class BranchIndexCacheController:
    """ BranchIndex Cache Controller

    The BranchIndex cache controller stores all the chunks in memory,
    allowing fast access from the mobile and desktop devices.

    """

    LOAD_CHUNK = 32

    def __init__(self, clientId: str):
        self._clientId = clientId
        self._webAppHandler = None

        #: This stores the cache of branchIndex data for the clients
        self._cache: Dict[int, BranchIndexEncodedChunk] = {}

        self._endpoint = PayloadEndpoint(clientBranchIndexUpdateFromServerFilt,
                                         self._processBranchIndexPayload)

    def setBranchIndexCacheHandler(self, handler):
        self._webAppHandler = handler

    @inlineCallbacks
    def start(self):
        yield self.reloadCache()

    def shutdown(self):
        self._tupleObservable = None

        self._endpoint.shutdown()
        self._endpoint = None

        self._cache = {}

    @inlineCallbacks
    def reloadCache(self):
        self._cache = {}

        offset = 0
        while True:
            logger.info(
                "Loading BranchIndexChunk %s to %s" % (offset, offset + self.LOAD_CHUNK))
            encodedChunkTuples: List[BranchIndexEncodedChunk] = (
                yield BranchIndexChunkLoadRpc.loadBranchIndexChunks(offset, self.LOAD_CHUNK)
            )

            if not encodedChunkTuples:
                break

            self._loadBranchIndexIntoCache(encodedChunkTuples)

            offset += self.LOAD_CHUNK

    @inlineCallbacks
    def _processBranchIndexPayload(self, payloadEnvelope: PayloadEnvelope, **kwargs):
        paylod = yield payloadEnvelope.decodePayloadDefer()
        branchIndexTuples: List[BranchIndexEncodedChunk] = paylod.tuples
        self._loadBranchIndexIntoCache(branchIndexTuples)

    def _loadBranchIndexIntoCache(self,
                                  encodedChunkTuples: List[BranchIndexEncodedChunk]):
        chunkKeysUpdated: List[str] = []

        for t in encodedChunkTuples:
            if (not t.chunkKey in self._cache or
                    self._cache[t.chunkKey].lastUpdate != t.lastUpdate):
                self._cache[t.chunkKey] = t
                chunkKeysUpdated.append(t.chunkKey)

        logger.debug("Received %s grids from server,"
                     " %s had changed ",
                     len(encodedChunkTuples), len(chunkKeysUpdated))

        self._webAppHandler.notifyOfBranchIndexUpdate(chunkKeysUpdated)

    def branchIndexChunk(self, chunkKey) -> BranchIndexEncodedChunk:
        return self._cache.get(chunkKey)

    def branchIndexKeys(self) -> List[int]:
        return list(self._cache)



"""Added edgeColor removed lu indexes

Peek Plugin Database Migration Script

Revision ID: da424430bf97
Revises: 473d5f6a97e8
Create Date: 2019-06-30 16:24:37.618354

"""

# revision identifiers, used by Alembic.
revision = 'da424430bf97'
down_revision = '473d5f6a97e8'
branch_labels = None
depends_on = None

import sqlalchemy as sa
from alembic import op


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index('idx_DispEllipse_fillColorId', table_name='DispEllipse', schema='pl_diagram')
    op.drop_index('idx_DispEllipse_lineColorId', table_name='DispEllipse', schema='pl_diagram')
    op.drop_index('idx_DispEllipse_lineStyleId', table_name='DispEllipse', schema='pl_diagram')
    op.drop_index('idx_DispPolygon_fillColorId', table_name='DispPolygon', schema='pl_diagram')
    op.drop_index('idx_DispPolygon_lineColorId', table_name='DispPolygon', schema='pl_diagram')
    op.drop_index('idx_DispPolygon_lineStyleId', table_name='DispPolygon', schema='pl_diagram')
    op.add_column('DispPolyline', sa.Column('edgeColorId', sa.Integer(), nullable=True), schema='pl_diagram')
    op.drop_index('idx_DispPolyline_lineColorId', table_name='DispPolyline', schema='pl_diagram')
    op.drop_index('idx_DispPolyline_lineStyleId', table_name='DispPolyline', schema='pl_diagram')
    op.create_foreign_key('DispGroupPointer_edgeColorId_fkey', 'DispPolyline', 'DispColor', ['edgeColorId'], ['id'], source_schema='pl_diagram', referent_schema='pl_diagram')
    op.drop_index('idx_DispText_colorId', table_name='DispText', schema='pl_diagram')
    op.drop_index('idx_DispText_styleId', table_name='DispText', schema='pl_diagram')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_index('idx_DispText_styleId', 'DispText', ['textStyleId'], unique=False, schema='pl_diagram')
    op.create_index('idx_DispText_colorId', 'DispText', ['colorId'], unique=False, schema='pl_diagram')
    op.drop_constraint('DispGroupPointer_edgeColorId_fkey', 'DispPolyline', schema='pl_diagram', type_='foreignkey')
    op.create_index('idx_DispPolyline_lineStyleId', 'DispPolyline', ['lineStyleId'], unique=False, schema='pl_diagram')
    op.create_index('idx_DispPolyline_lineColorId', 'DispPolyline', ['lineColorId'], unique=False, schema='pl_diagram')
    op.drop_column('DispPolyline', 'edgeColorId', schema='pl_diagram')
    op.create_index('idx_DispPolygon_lineStyleId', 'DispPolygon', ['lineStyleId'], unique=False, schema='pl_diagram')
    op.create_index('idx_DispPolygon_lineColorId', 'DispPolygon', ['lineColorId'], unique=False, schema='pl_diagram')
    op.create_index('idx_DispPolygon_fillColorId', 'DispPolygon', ['fillColorId'], unique=False, schema='pl_diagram')
    op.create_index('idx_DispEllipse_lineStyleId', 'DispEllipse', ['lineStyleId'], unique=False, schema='pl_diagram')
    op.create_index('idx_DispEllipse_lineColorId', 'DispEllipse', ['lineColorId'], unique=False, schema='pl_diagram')
    op.create_index('idx_DispEllipse_fillColorId', 'DispEllipse', ['fillColorId'], unique=False, schema='pl_diagram')
    # ### end Alembic commands ###